#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "sym_table.h"
#include "mylc3as.h"
#include "parser.h"

FILE* g_outputfile;  

void write_word_to_file(int word)
{
  fputc((word >> 8) & 0xff, g_outputfile);
  fputc(word & 0xff, g_outputfile);
}

int add_field(int bits, int new_bits, int hi_bit, int lo_bit)
{
  // Check for reversed operands
  if (lo_bit > hi_bit) {
    return add_field(bits, new_bits, lo_bit, hi_bit);
  }

  int mask = ~0;
  mask = mask << (hi_bit - lo_bit + 1);
  mask = ~mask;
  mask = mask << lo_bit;
  
  return (mask & (new_bits << lo_bit)) | (~mask & bits);
}

int next_address(insn_t* insn_ptr, int current_address)
{
  // insert your code below
  switch (insn_ptr->opcode) {
  case ORIG:
    return insn_ptr->offset_immediate;
  case BLKW:
    /// insert your code here
  case STRINGZ:
    /// insert your code here
  case UNDEF:                      // label only lines
    return current_address;
  default:
    return current_address + 1;
  }
  return current_address;
}

// Pass #1: insert labels into symbol table
void pass_one(insn_t* insn_ptr)
{
  int current_address = 0;
  while (insn_ptr != NULL) {

    /// insert your code here
    
    current_address = next_address(insn_ptr, current_address);
    insn_ptr = insn_ptr->next;
  }
}

// Pass #2: lookup labels in symbol table, fill in offset_immediate fields if necessary
void pass_two(insn_t* insn_ptr)
{
  /// insert your code here basic structure the same as pass_one).
}

// Pass #3: generate instructions
void pass_three(insn_t* insn_ptr) 
{
  while (insn_ptr != NULL) {
    switch (insn_ptr->opcode) {
    case ADD:                                        /* ADD */
    case AND:                                        /* AND */
    case MUL:                                        /* MUL */
      {
        int insn = 0;
        insn = add_field(insn, insn_ptr->opcode, 15, 12); // OPCODE
        insn = add_field(insn, insn_ptr->first_reg, 11, 9); // DR
        insn = add_field(insn, insn_ptr->second_reg, 8, 6); // SR1
        if (insn_ptr->third_reg == -1) {
          insn = add_field(insn, 0x1, 5, 5); // Imm5
          insn = add_field(insn, insn_ptr->offset_immediate, 4, 0); // Imm5
        } else {
          insn = add_field(insn, insn_ptr->third_reg, 2, 0); // SR2
        }
        write_word_to_file(insn);
      }
      break;
    case SUB:                                         /* SUB */
      /// insert your code here
      break;
    case BR:                                         /* BR */
      /// insert your code here
      break;
    case JMP_RET:                                    /* JMP/RET */
      /// insert your code here
      break;
    case JSR_JSRR:                                   /* JSR/JSRR */
      /// insert your code here
      break;
    case LD:                                         /* LD */
    case ST:                                         /* ST */
    case LDI:                                        /* LDI */
    case STI:                                        /* STI */
    case LEA:                                        /* LEA */
      /// insert your code here
      break;
    case LDR:                                        /* LDR */
    case STR:                                        /* STR */
      /// insert your code here
      break;
    case NOT:                                        /* NOT */
      /// insert your code here
      break;
    case RTI:                                        /* RTI */
      /// insert your code here
      break;
    case TRAP:                                       /* TRAP */
      /// insert your code here
      break;
    case FILL:                                       /* .FILL */
    case ORIG:                                       /* .ORIG */
      /// insert your code here
      break;
    case BLKW:                                       /* .BLKW */
      /// insert your code here
      break;
    case STRINGZ:                                    /* .STRINGZ */
      /// insert your code here
      break;
    case UNDEF:                                      /* label-only lines */
      // no instruction, so do nothing
      break;

    default:
      printf("ERROR: Unexpected opcode (%d)\n", insn_ptr->opcode);
      exit(1);
    }
    insn_ptr = insn_ptr->next;
  }
}

int main(int argc, char **argv) 
{
  FILE* f;
  insn_t* starting_insn_ptr = NULL;

  if (argc != 3) {
    printf("Usage: %s <asm-file> <obj-file>\n", argv[0]);
    return 1;
  }

  // open file for reading
  f = fopen(argv[1], "r");
  if (f == NULL) {
    printf("Error opening file %s for reading\n", argv[1]);
    return 1;
  }

  g_outputfile = fopen(argv[2], "wb");
  if (g_outputfile == NULL) {
    printf("Error opening file %s for writing\n", argv[2]);
    return 1;
  }

  starting_insn_ptr = parse(f); // from parser.c
  pass_one(starting_insn_ptr);
  pass_two(starting_insn_ptr);
  pass_three(starting_insn_ptr);

  return 0;
}

